/* 
   $Id: sgworkspace.c,v 1.10 2001/12/25 05:15:36 sho Exp $

   Copyright (C) 2001,2002 AXE,Inc. <sg@axe-inc.co.jp>

   This file is part of sgwm.

   sgwm is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   sgwm is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with sgwm; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "sgwm.h"
#include <gdk/gdkx.h>
#include <gtk/gtksignal.h>
#include "sgwmwidget.h"
#include "sgworkspace.h"
#include "sgwindow.h"

static void sgworkspace_class_init(SGWorkSpaceClass *klass);
static void sgworkspace_init(SGWorkSpace *win);
static GtkContainerClass *parent_class = NULL;

GtkType sgworkspace_get_type(void)
{
    static GtkType sgworkspace_type = 0;

    if (sgworkspace_type == 0) {
	GtkTypeInfo sgworkspace_info = {
	    "SGWorkSpace",
	    sizeof(SGWorkSpace),
	    sizeof(SGWorkSpaceClass),
	    (GtkClassInitFunc)sgworkspace_class_init,
	    (GtkObjectInitFunc)sgworkspace_init,
	    NULL,
	    NULL,
	    NULL
	};
	sgworkspace_type = gtk_type_unique(gtk_container_get_type(),
	    &sgworkspace_info);
    }

    return sgworkspace_type;
}

static void sgworkspace_focus_in(SGWorkSpace *ws, XEvent *ev)
{
    if (ws->children == NULL) return;
    if (ev->xfocus.mode != NotifyNormal
      || ev->xfocus.detail != NotifyNonlinearVirtual) return;
    sgwmwidget_do_activate_workspace(ws->sgwm,ws->no,FALSE,FALSE);
}

static GdkFilterReturn sgworkspace_filter(GdkXEvent *xevent,
 GdkEvent *event, gpointer data)
{
    XEvent *ev = (XEvent *)xevent;

    if (ev->type == FocusIn) {
	sgworkspace_focus_in((SGWorkSpace *)data,ev);
	return GDK_FILTER_REMOVE;
    }
    return GDK_FILTER_CONTINUE;
}

static void sgworkspace_bound_size(SGWorkSpace *ws, gint *width, gint *height)
{
    GList *cur;
    SGWindow *sgwin;
    int minx, miny;
    int maxx, maxy;
    int f = 0;

    minx = miny = G_MAXINT;
    maxx = maxy = G_MININT;
    for (cur = ws->children;cur != NULL;cur = cur->next) {
	sgwin = (SGWindow *)cur->data;
	if (sgwin != NULL && !SGWINDOW_DESTROYED(sgwin)
	   && GTK_WIDGET_REALIZED(sgwin)) {
	    gint x,y,w,h,bw;

	    f = 1;
	    gdk_error_trap_push();
	    sgwindow_get_geometry(sgwin,&x,&y,&w,&h,&bw);
	    gdk_flush();
	    if (!gdk_error_trap_pop()) {
		minx = MIN(minx,x);
		miny = MIN(miny,y);
		x += w;
		y += h;
		maxx = MAX(maxx,x);
		maxy = MAX(maxy,y);
	    }
	}
    }
    if (f) {
	*width = maxx-minx;
	*height = maxy-miny;
    } else {
	*width = 1;
	*height = 1;
    }
}

static void sgworkspace_state_changed(GtkWidget *widget, GtkStateType oldstate)
{
    if (GTK_WIDGET_REALIZED(widget)) {
	gtk_style_set_background(widget->style, widget->window, widget->state);
	gdk_window_clear(widget->window);
    }
}

static void sgworkspace_realize(GtkWidget *widget)
{
    GdkWindowAttr attributes;
    gint attributes_mask;
    gint bw;
    SGWorkSpace *ws;
    gint w,h;

    ws = SGWORKSPACE(widget);
    GTK_WIDGET_SET_FLAGS(widget,GTK_REALIZED);

    bw = GTK_CONTAINER(widget)->border_width;
    attributes.x = widget->allocation.x+bw;
    attributes.y = widget->allocation.y+bw;
    attributes.width = widget->allocation.width-bw*2;
    attributes.height = widget->allocation.height-bw*2;
    attributes.window_type = GDK_WINDOW_CHILD;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.visual = gtk_widget_get_visual(widget);
    attributes.colormap = gtk_widget_get_colormap(widget);
    attributes.event_mask = 0;
    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
    widget->window = gdk_window_new(gtk_widget_get_parent_window(widget),
      &attributes,attributes_mask);
    gdk_window_set_user_data(widget->window, widget);
    widget->style = gtk_style_attach(widget->style, widget->window);
    gtk_style_set_background(widget->style, widget->window, widget->state);
    gdk_window_add_filter(widget->window,sgworkspace_filter,
      (gpointer)ws);
    XSelectInput(GDK_DISPLAY(), GDK_WINDOW_XWINDOW(widget->window),
	FocusChangeMask | EnterWindowMask);

    attributes.x = widget->style->klass->xthickness;
    attributes.y = widget->style->klass->ythickness;
    attributes.width -= attributes.x*2;
    attributes.height -= attributes.y*2;
    ws->view_window = gdk_window_new(widget->window,
      &attributes,attributes_mask);
    gtk_style_set_background(widget->style, ws->view_window, GTK_STATE_NORMAL);

    sgworkspace_bound_size(ws,&w,&h);
    attributes.width = G_MAXSHORT;
    attributes.height = G_MAXSHORT;
    attributes.x = attributes.y = 0;
    attributes.event_mask = 0;
    ws->bin_window = gdk_window_new(ws->view_window,
      &attributes,attributes_mask);
    gdk_window_add_filter(ws->bin_window,sgwmwidget_filter,
      (gpointer)ws->sgwm);
    XSelectInput(GDK_DISPLAY(), GDK_WINDOW_XWINDOW(ws->bin_window),
	SubstructureRedirectMask|SubstructureNotifyMask);
    gtk_style_set_background(widget->style, ws->bin_window, GTK_STATE_NORMAL);
    {
	ws->cover = XCreateWindow(GDK_DISPLAY(),GDK_WINDOW_XWINDOW(ws->view_window),
	  0,0,G_MAXSHORT,G_MAXSHORT,0,CopyFromParent,InputOutput,CopyFromParent,
	  0,NULL);
    }
}

void sgworkspace_cover(GtkWidget *widget)
{
    SGWorkSpace *ws = SGWORKSPACE(widget);

    if (ws->coverCount++ <= 0) {
	XMapRaised(GDK_DISPLAY(),ws->cover);
    }
}

void sgworkspace_uncover(GtkWidget *widget)
{
    SGWorkSpace *ws = SGWORKSPACE(widget);

    if (--ws->coverCount <= 0) {
	XUnmapWindow(GDK_DISPLAY(),ws->cover);
	ws->coverCount = 0;
    }
}

static void sgworkspace_map(GtkWidget *widget)
{
    SGWorkSpace *ws;

    GTK_WIDGET_SET_FLAGS(widget,GTK_MAPPED);
    ws = SGWORKSPACE(widget);
    gdk_window_show(widget->window);
    gdk_window_show(ws->view_window);
    gdk_window_show(ws->bin_window);
}

static void sgworkspace_do_calculate_child_size(SGWorkSpace *ws,
  SGWindow *sgwin, gint rbw, GtkAllocation *alloc, GtkAllocation *newalloc)
{
    XSizeHints *hints;
    GtkWidget *widget = GTK_WIDGET(sgwin);

    if (sgwin->type == SGWINDOW_NUNOME) {
	newalloc->x = 0;
	newalloc->y = 0;
	return;
    }
    if (GTK_WIDGET_REALIZED(widget) && !SGWINDOW_DESTROYED(sgwin)) {
	if ((hints = XAllocSizeHints()) != NULL) {
	    if (sgwindow_get_normal_hints(sgwin,hints)) {
		if (!(hints->flags & USPosition)) {
		    newalloc->x = 0;
		    newalloc->y = 0;
		}
		if (!(hints->flags & USSize)) {
		    gint mw,mh;
		    gboolean fw = FALSE, fh = FALSE;

		    if ((hints->flags & PMaxSize)) {
			if (hints->max_width+newalloc->x+rbw*2 <=
			   alloc->width) {
			    newalloc->width = hints->max_width;
			    fw = TRUE;
			}
			if (hints->max_height+newalloc->y+rbw*2 <=
			   alloc->height) {
			    newalloc->height = hints->max_height;
			    fh = TRUE;
			}
		    }
		    if (hints->flags & PBaseSize) {
			mw = hints->base_width;
			mh = hints->base_height;
		    } else if (hints->flags & PMinSize) {
			mw = hints->min_width;
			mh = hints->min_height;
		    } else {
			mw = 1;
			mh = 1;
		    }
		    if (mw+newalloc->x+rbw*2 > alloc->width) {
			fw = TRUE;
		    }
		    if (mh+newalloc->y+rbw*2 > alloc->height) {
			fh = TRUE;
		    }
		    if (!fw) {
			newalloc->width = alloc->width-rbw*2;
			if ((hints->flags & PResizeInc)) {
			    newalloc->width = (newalloc->width-mw)/hints->width_inc;
			    newalloc->width = hints->width_inc*newalloc->width+mw;
			}
		    }
		    if (!fh) {
			newalloc->height = alloc->height-rbw*2;
			if ((hints->flags & PResizeInc)) {
			    newalloc->height = (newalloc->height-mh)/
			                         hints->height_inc;
			    newalloc->height = hints->height_inc*newalloc->height+mh;
			}
		    }
		    if (hints->flags & PAspect) {
			/* Notice: PAspect is not supported. */
		    }
		}
	    } else {
		*newalloc = *alloc;
	    }
	    XFree(hints);
	} else {
	    *newalloc = *alloc;
	}
    }
}

void sgworkspace_calculate_child_size(SGWorkSpace *ws,
  SGWindow *sgwin, GtkAllocation *newalloc)
{
    GtkWidget *widget = GTK_WIDGET(ws);
    GtkAllocation alloc;
    gint tx,ty,tw,th,tbw;
    gint bw;

    if (GTK_WIDGET_REALIZED(widget)) {
	alloc = widget->allocation;
	bw = GTK_CONTAINER(widget)->border_width;
	alloc.x = 0;
	alloc.y = 0;
	alloc.width -= (bw+alloc.x)*2;
	alloc.height -= (bw+alloc.y)*2;
	sgwindow_get_geometry(sgwin,&tx,&ty,&tw,&th,&tbw);
	sgworkspace_do_calculate_child_size(ws,sgwin,tbw,&alloc,newalloc);
    }
}

static void sgworkspace_do_resize_children(SGWorkSpace *ws,
 GtkAllocation *alloc)
{
    GList *cur;
    SGWindow *sgwin;
    gint tx,ty,tw,th,tbw;
    GtkAllocation newalloc;
    
    for (cur = ws->children;cur != NULL;cur = cur->next) {
	sgwin = (SGWindow *)cur->data;
	sgwindow_get_geometry(sgwin,&tx,&ty,&tw,&th,&tbw);
	newalloc.x = tx;
	newalloc.y = ty;
	newalloc.width = tw;
	newalloc.height = th;
	sgworkspace_do_calculate_child_size(ws,sgwin,tbw,alloc,&newalloc);
	sgwindow_move_resize(sgwin,newalloc.x,newalloc.y,newalloc.width,
	   newalloc.height);
    }
}

static void sgworkspace_size_allocate(GtkWidget *widget,
	GtkAllocation *allocation)
{
    SGWorkSpace *ws;
    gint bw;
    GtkAllocation alloc;
    gint w,h;

    ws = SGWORKSPACE(widget);
    if (!ws->childrenChanged
      && widget->allocation.x == allocation->x
      && widget->allocation.y == allocation->y
      && widget->allocation.width == allocation->width
      && widget->allocation.height == allocation->height) {
	return;
    }
    ws->childrenChanged = FALSE;
    alloc = widget->allocation = *allocation;
    if (GTK_WIDGET_REALIZED(widget)) {
	gint hval,vval,xt,yt;

	bw = GTK_CONTAINER(widget)->border_width;
	xt = widget->style->klass->xthickness;
	yt = widget->style->klass->ythickness;
	alloc.x += bw;
	alloc.y += bw;
	if (alloc.width > bw*2) {
	    alloc.width -= bw*2;
	} else {
	    alloc.width = 1;
	}
	if (alloc.height > bw*2) {
	    alloc.height -= bw*2;
	} else {
	    alloc.height = 1;
	}
	gdk_window_move_resize(widget->window,
	  alloc.x,
	  alloc.y,
	  alloc.width,
	  alloc.height);
	alloc.x = xt;
	alloc.y = yt;
	if (alloc.width > xt*2) {
	    alloc.width -= xt*2;
	} else {
	    alloc.width = 1;
	}
	if (alloc.height > yt*2) {
	    alloc.height -= yt*2;
	} else {
	    alloc.height = 1;
	}
	gdk_window_move_resize(ws->view_window,
	  alloc.x,
	  alloc.y,
	  alloc.width,
	  alloc.height);
	alloc.x = alloc.y = 0;
	sgworkspace_do_resize_children(ws,&alloc);
	sgworkspace_bound_size(ws,&w,&h);
	w = MAX(w,alloc.width);
	h = MAX(h,alloc.height);
	if (ws->hadj != NULL) {
	    ws->hadj->page_size = alloc.width;
	    ws->hadj->page_increment = alloc.width/2;
	    ws->hadj->step_increment = 10;
	    ws->hadj->lower = 0;
	    ws->hadj->upper = w;
	    gtk_adjustment_changed(ws->hadj);
	    hval = CLAMP(ws->hadj->value,0,ws->hadj->upper - ws->hadj->page_size);
	    if (ws->hadj->value != hval) {
		ws->hadj->value = hval;
		gtk_adjustment_value_changed(ws->hadj);
	    }
	}
	if (ws->vadj != NULL) {
	    ws->vadj->page_size = alloc.height;
	    ws->vadj->page_increment = alloc.height/2;
	    ws->vadj->step_increment = 10;
	    ws->vadj->lower = 0;
	    ws->vadj->upper = h;
	    gtk_adjustment_changed(ws->vadj);
	    vval = CLAMP(ws->vadj->value,0,ws->vadj->upper - ws->vadj->page_size);
	    if (ws->vadj->value != vval) {
		ws->vadj->value = vval;
		gtk_adjustment_value_changed(ws->vadj);
	    }
	}
    }
}

static void sgworkspace_disable_scrollbar(SGWorkSpace *ws)
{

    if (ws->hadj != NULL) {
	ws->hadj->page_size = 1;
	ws->hadj->lower = 0;
	ws->hadj->upper = 0;
	gtk_adjustment_changed(ws->hadj);
    }
    if (ws->vadj != NULL) {
	ws->vadj->page_size = 1;
	ws->hadj->lower = 0;
	ws->vadj->upper = 0;
	gtk_adjustment_changed(ws->vadj);
    }
}

void sgworkspace_add(GtkContainer *con, GtkWidget *sgwin)
{
    SGWorkSpace *ws;
    GtkWidget *widget;

    ws = SGWORKSPACE(con);
    widget = GTK_WIDGET(con);
    if (sgwin->parent != widget) {
	gdk_error_trap_push();
	gtk_widget_set_parent_window(sgwin,ws->bin_window);
	gtk_widget_set_parent(sgwin,widget);
	gtk_widget_realize(sgwin);
	gtk_widget_show(sgwin);
	gdk_flush();
	if (gdk_error_trap_pop()) {
	    ;
	}
	ws->children = g_list_append(ws->children, (gpointer)sgwin);
	ws->childrenChanged = TRUE;
    }
}

void sgworkspace_remove(GtkContainer *con, GtkWidget *sgwin)
{
    SGWorkSpace *ws;

    ws = SGWORKSPACE(con);
    if (g_list_find(ws->children, (gpointer)sgwin) == NULL) return;
    ws->children = g_list_remove(ws->children, (gpointer)sgwin);
    gdk_error_trap_push();
    XGrabServer(GDK_DISPLAY());
    XSelectInput(GDK_DISPLAY(), GDK_WINDOW_XWINDOW(ws->bin_window),0);
    gtk_widget_hide(sgwin);
    XSelectInput(GDK_DISPLAY(), GDK_WINDOW_XWINDOW(ws->bin_window),
	SubstructureRedirectMask|SubstructureNotifyMask);
    XUngrabServer(GDK_DISPLAY());
    gtk_widget_unparent(sgwin);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	;
    }
    ws->childrenChanged = TRUE;
    sgworkspace_disable_scrollbar(ws);
}

gboolean sgworkspace_in(SGWorkSpace *sgws, int x, int y)
{
    int rx,ry;
    GtkWidget *widget = GTK_WIDGET(sgws);
    Window rwin;

    if (!GTK_WIDGET_REALIZED(sgws)) return FALSE;
    XTranslateCoordinates(GDK_DISPLAY(),GDK_ROOT_WINDOW(),
      GDK_WINDOW_XWINDOW(widget->window),
      x,y,&rx,&ry,&rwin);
    return (rx >= 0 && rx < widget->allocation.width
           && ry >= 0 && ry < widget->allocation.height);
}

static void sgworkspace_adj_changed(GtkAdjustment *adj, gpointer *p)
{
}

static void sgworkspace_adj_value_changed(GtkAdjustment *adj, gpointer *p)
{
    gint x,y;
    SGWorkSpace *ws = SGWORKSPACE(p);

    x = y = 0;
    if (ws->hadj != NULL) {
	x = ws->hadj->lower - ws->hadj->value;
    }
    if (ws->vadj != NULL) {
	y = ws->vadj->lower - ws->vadj->value;
    }
    if (GTK_WIDGET_REALIZED(ws)) {
	gdk_window_move(ws->bin_window,x,y);
    }
}

static void sgworkspace_set_hadj(SGWorkSpace *ws, GtkAdjustment *hadj)
{
    ws->hadj = hadj;
    if (hadj != NULL) {
	gtk_object_ref(GTK_OBJECT(hadj));
	gtk_signal_connect (GTK_OBJECT(hadj), "changed",
	  (GtkSignalFunc)sgworkspace_adj_changed, (gpointer)ws);
	gtk_signal_connect (GTK_OBJECT(hadj), "value_changed",
	  (GtkSignalFunc)sgworkspace_adj_value_changed, (gpointer)ws);
    }
    sgworkspace_adj_changed(hadj,(gpointer)ws);
}

static void sgworkspace_set_vadj(SGWorkSpace *ws, GtkAdjustment *vadj)
{
    ws->vadj = vadj;
    if (vadj != NULL) {
	gtk_object_ref(GTK_OBJECT(vadj));
	gtk_signal_connect (GTK_OBJECT(vadj), "changed",
	  (GtkSignalFunc)sgworkspace_adj_changed, (gpointer)ws);
	gtk_signal_connect (GTK_OBJECT(vadj), "value_changed",
	  (GtkSignalFunc)sgworkspace_adj_value_changed, (gpointer)ws);
    }
    sgworkspace_adj_changed(vadj,(gpointer)ws);
}

static void sgworkspace_set_adjs(SGWorkSpace *ws, GtkAdjustment *hadj,
  GtkAdjustment *vadj)
{
    if (ws->hadj != hadj) {
	sgworkspace_set_hadj(ws,hadj);
    }
    if (ws->vadj != vadj) {
	sgworkspace_set_vadj(ws,vadj);
    }
}

static gint sgworkspace_enter_notify_event(GtkWidget *widget, GdkEventCrossing *ev)
{
    SGWorkSpace *ws = SGWORKSPACE(widget);

    if (ws->children == NULL) return TRUE;
    if (ev->mode == GDK_CROSSING_UNGRAB) return TRUE;
    sgwmwidget_do_activate_workspace(ws->sgwm,ws->no,FALSE,TRUE);
    return TRUE;
}

static void sgworkspace_style_set(GtkWidget *widget, GtkStyle *p_style)
{
    SGWorkSpace *ws = SGWORKSPACE(widget);

    if (ws->view_window != NULL) {
	gtk_style_set_background(widget->style,
	  ws->view_window, GTK_STATE_NORMAL);
    }
    if (ws->bin_window != NULL) {
	gtk_style_set_background(widget->style,
	  ws->bin_window, GTK_STATE_NORMAL);
    }
}

static void sgworkspace_class_init(SGWorkSpaceClass *klass)
{
    GtkWidgetClass *widget_class;
    GtkContainerClass *container_class;
    GtkObjectClass *object_class;

    parent_class = gtk_type_class(gtk_container_get_type());
    widget_class = (GtkWidgetClass *)klass;
    object_class = (GtkObjectClass *)klass;
    container_class = (GtkContainerClass *)klass;

    widget_class->realize = sgworkspace_realize;
    widget_class->map = sgworkspace_map;
    widget_class->state_changed = sgworkspace_state_changed;
    widget_class->size_allocate = sgworkspace_size_allocate;
    widget_class->enter_notify_event = sgworkspace_enter_notify_event;
    widget_class->set_scroll_adjustments_signal =
	gtk_signal_new ("set_scroll_adjustments",
                    GTK_RUN_LAST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SGWorkSpaceClass, set_scroll_adjustments),
                    gtk_marshal_NONE__POINTER_POINTER,
                    GTK_TYPE_NONE, 2, GTK_TYPE_ADJUSTMENT, GTK_TYPE_ADJUSTMENT);
    widget_class->style_set = sgworkspace_style_set;
    klass->set_scroll_adjustments = sgworkspace_set_adjs;


    container_class->add = sgworkspace_add;
    container_class->remove = sgworkspace_remove;
}

static void sgworkspace_init(SGWorkSpace *ws)
{
    ws->view_window = NULL;
    ws->bin_window = NULL;
    ws->children = NULL;
    ws->disabled = FALSE;
    ws->nunome = FALSE;
    ws->childrenChanged = FALSE;
    ws->cover = None;
    ws->coverCount = 0;
}

GtkWidget *sgworkspace_new(SGWMWidget *sgwm, GtkAdjustment *hadj,
  GtkAdjustment *vadj, gint no)
{
    GtkWidget *widget;
    SGWorkSpace *ws;

    widget  = GTK_WIDGET(gtk_type_new(sgworkspace_get_type()));
    ws = SGWORKSPACE(widget);
    ws->sgwm = sgwm;
    ws->hadj = ws->vadj = NULL;
    ws->no = no;
    sgworkspace_set_adjs(ws,hadj,vadj);
    return widget;
}

void sgworkspace_warp_pointer(SGWorkSpace *sgws)
{
    GtkWidget *widget;
    gint x,y;
    GdkModifierType mask;

    if (sgws == NULL) return;
    widget = GTK_WIDGET(sgws);
    if (widget->window == NULL) return;
    gdk_error_trap_push();
    gdk_flush();
    gdk_window_get_pointer(widget->window,&x,&y,&mask);
    if (widget->allocation.width < x || x < 0
       || widget->allocation.height < y || y < 0) {
	XWarpPointer(GDK_DISPLAY(),None,GDK_WINDOW_XWINDOW(widget->window),
	  0,0,0,0,widget->allocation.width/2,widget->allocation.height/2);
    }
    gdk_flush();
    if (gdk_error_trap_pop()) {
	;
    }
}

void sgworkspace_send_client_message(GtkWidget *widget, GdkEvent *ev)
{
    GList *cur;
    SGWorkSpace *ws = SGWORKSPACE(widget);
    Window ret_root, ret_parent, *ret_children;
    unsigned int n;
    int i;

    for (cur = ws->children;cur != NULL;cur = cur->next) {
	if (ws->bin_window != NULL) {

	    gdk_error_trap_push();
	    if (XQueryTree(GDK_DISPLAY(),GDK_WINDOW_XWINDOW(ws->bin_window),
	         &ret_root,&ret_parent,&ret_children,&n) == True) {
		    for (i = 0;i < n;i++) {
			gdk_event_send_client_message(ev,ret_children[i]);
		    }
	    }
	    XFree(ret_children);
	    gdk_flush();
	    if (gdk_error_trap_pop()) {
		;
	    }
	}
    }
}
